#pragma once

// reimplementation of std::integral_constant
template<class _Type, _Type _Value>
struct MphIntegralConstant {
	using value_type = _Type;
	using type = MphIntegralConstant;
	
	static constexpr _Type value = _Value;

	constexpr operator value_type() const noexcept { return value; }
	constexpr value_type operator()() const noexcept { return value; }
};

// reimplementation of std::true_type and std::false_type
using MphTrueType  = MphIntegralConstant<bool, true>;
using MphFalseType = MphIntegralConstant<bool, false>;

// reimplementation of std::is_same
template<class _Type1, class _Type2>
struct MphIsSame : MphFalseType {};

template<class _Type>
struct MphIsSame<_Type, _Type> : MphTrueType {};

// reimplementation of std::same_as
template<class _Type1, class _Type2>
concept MphSameAs = MphIsSame<_Type1, _Type2>::value
                 && MphIsSame<_Type2, _Type1>::value;

// reimplementation of std::is_pointer
template<class _Type>
struct MphIsPointer : MphFalseType {};

template<class _Type>
struct MphIsPointer<_Type*> : MphTrueType {};

template<class _Type>
struct MphIsPointer<_Type* const> : MphTrueType {};

template<class _Type>
struct MphIsPointer<_Type* volatile> : MphTrueType {};

template<class _Type>
struct MphIsPointer<_Type* const volatile> : MphTrueType {};

// checks if something can be static casted to something else
template<class _From, class _To>
concept MphStaticCastableTo = requires(_From&& val) {
	static_cast<_To>(val);
};
